@@ -15,7 +15,7 @@ static void usage(void) {
 		"\n"
 		"Commands:\n"
 		"  reload            Reload the configuration file\n"
-		"  switch <profile>  Switch to another profile");
+		"  switch <profile>  Switch to another profile\n");
 }
 
 static long handle_call_done(VarlinkConnection *connection, const char *error,
@@ -25,6 +25,8 @@ static long handle_call_done(VarlinkConnection *connection, const char *error,
 			fprintf(stderr, "Profile not found\n");
 		} else if (strcmp(error, "fr.emersion.kanshi.ProfileNotMatched") == 0) {
 			fprintf(stderr, "Profile does not match the current output configuration\n");
+		} else if (strcmp(error, "fr.emersion.kanshi.ProfileNotApplied") == 0) {
+			fprintf(stderr, "Profile could not be applied by the compositor\n");
 		} else {
 			fprintf(stderr, "Error: %s\n", error);
 		}
@@ -47,7 +49,7 @@ static int set_blocking(int fd) {
 	return 0;
 }
 
-int wait_for_event(VarlinkConnection *connection) {
+static int wait_for_event(VarlinkConnection *connection) {
 	int fd = varlink_connection_get_fd(connection);
 	if (set_blocking(fd) != 0) {
 		return -1;
