@@ -20,7 +20,8 @@
 
 #define HEADS_MAX 64
 
-static bool match_and_apply(struct kanshi_state *state);
+static bool match_and_apply(struct kanshi_state *state,
+	kanshi_apply_done_func callback, void *data);
 
 static bool match_profile_output(struct kanshi_profile_output *output,
 		struct kanshi_head *head) {
@@ -150,6 +151,9 @@ static void config_handle_succeeded(void *data,
 	if (profile == state->pending_profile) {
 		state->pending_profile = NULL;
 	}
+	if (pending->callback != NULL) {
+		pending->callback(pending->callback_data, true);
+	}
 	free(pending);
 }
 
@@ -162,6 +166,9 @@ static void config_handle_failed(void *data,
 	if (pending->profile == pending->state->pending_profile) {
 		pending->state->pending_profile = NULL;
 	}
+	if (pending->callback != NULL) {
+		pending->callback(pending->callback_data, false);
+	}
 	free(pending);
 }
 
@@ -178,7 +185,10 @@ static void config_handle_cancelled(void *data,
 	if (pending->serial != pending->state->serial) {
 		// We've already received a new serial, try re-applying the profile
 		// immediately
-		match_and_apply(pending->state);
+		match_and_apply(pending->state, NULL, NULL);
+	}
+	if (pending->callback != NULL) {
+		pending->callback(pending->callback_data, false);
 	}
 	free(pending);
 }
@@ -230,11 +240,14 @@ static struct kanshi_mode *match_mode(struct kanshi_head *head,
 	return last_match;
 }
 
-static void apply_profile(struct kanshi_state *state,
-		struct kanshi_profile *profile,
-		struct kanshi_profile_output **matches) {
+static bool apply_profile(struct kanshi_state *state,
+		struct kanshi_profile *profile, struct kanshi_profile_output **matches,
+		kanshi_apply_done_func callback, void *data) {
 	if (state->pending_profile == profile || state->current_profile == profile) {
-		return;
+		if (callback != NULL) {
+			callback(data, true);
+		}
+		return true;
 	}
 
 	fprintf(stderr, "applying profile '%s'\n", profile->name);
@@ -243,6 +256,8 @@ static void apply_profile(struct kanshi_state *state,
 	pending->serial = state->serial;
 	pending->state = state;
 	pending->profile = profile;
+	pending->callback = callback;
+	pending->callback_data = data;
 	state->pending_profile = profile;
 
 	struct zwlr_output_configuration_v1 *config =
@@ -272,20 +287,25 @@ static void apply_profile(struct kanshi_state *state,
 		struct zwlr_output_configuration_head_v1 *config_head =
 			zwlr_output_configuration_v1_enable_head(config, head->wlr_head);
 		if (profile_output->fields & KANSHI_OUTPUT_MODE) {
-			// TODO: support custom modes
-			struct kanshi_mode *mode = match_mode(head,
-				profile_output->mode.width, profile_output->mode.height,
-				profile_output->mode.refresh);
-			if (mode == NULL) {
-				fprintf(stderr,
-					"output '%s' doesn't support mode '%dx%d@%fHz'\n",
-					head->name,
+			if (profile_output->mode.custom) {
+				fprintf(stderr, "applying the custom mode %s\n", profile_output->name);
+				zwlr_output_configuration_head_v1_set_custom_mode(config_head,
+					profile_output->mode.width, profile_output->mode.height, profile_output->mode.refresh);
+			} else {
+				struct kanshi_mode *mode = match_mode(head,
 					profile_output->mode.width, profile_output->mode.height,
-					(float)profile_output->mode.refresh / 1000);
-				goto error;
+					profile_output->mode.refresh);
+				if (mode == NULL) {
+					fprintf(stderr,
+						"output '%s' doesn't support mode '%dx%d@%fHz'\n",
+						head->name,
+						profile_output->mode.width, profile_output->mode.height,
+						(float)profile_output->mode.refresh / 1000);
+					goto error;
+				}
+				zwlr_output_configuration_head_v1_set_mode(config_head,
+						mode->wlr_mode);
 			}
-			zwlr_output_configuration_head_v1_set_mode(config_head,
-				mode->wlr_mode);
 		}
 		if (profile_output->fields & KANSHI_OUTPUT_POSITION) {
 			zwlr_output_configuration_head_v1_set_position(config_head,
@@ -306,10 +326,12 @@ static void apply_profile(struct kanshi_state *state,
 	}
 
 	zwlr_output_configuration_v1_apply(config);
-	return;
+	return true;
 
 error:
+	free(pending);
 	zwlr_output_configuration_v1_destroy(config);
+	return false;
 }
 
 
@@ -502,38 +524,53 @@ static void output_manager_handle_head(void *data,
 	zwlr_output_head_v1_add_listener(wlr_head, &head_listener, head);
 }
 
-static bool match_and_apply(struct kanshi_state *state) {
+static bool match_and_apply(struct kanshi_state *state,
+		kanshi_apply_done_func callback, void *data) {
 	assert(wl_list_length(&state->heads) <= HEADS_MAX);
 	// matches[i] gives the kanshi_profile_output for the i-th head
 	struct kanshi_profile_output *matches[HEADS_MAX];
 	if (state->current_profile != NULL &&
 			match_profile(state, state->current_profile, matches)) {
-		return true; // keep the current profile if it still matches
+		// keep the current profile if it still matches
+		if (callback != NULL) {
+			callback(data, true);
+		}
+		return true;
 	}
 	struct kanshi_profile *profile = match(state, matches);
 	if (profile != NULL) {
-		apply_profile(state, profile, matches);
-		return true;
+		if (apply_profile(state, profile, matches, callback, data)) {
+			return true;
+		}
+	} else {
+		fprintf(stderr, "no profile matched\n");
 	}
-	fprintf(stderr, "no profile matched\n");
+
+	// If a profile failed to match or apply, forget the current profile.
+	// This is necessary to make the following scenario work as expected:
+	// have a profile for DP-1 and DP-2, disconnect DP-2, reconnect DP-2.
+	// No profile will be matched in the intermediary state where only DP-1 is
+	// connected, however the profile needs to be re-applied for the final
+	// state.
+	state->current_profile = NULL;
 	return false;
 }
 
-bool kanshi_switch(struct kanshi_state *state, struct kanshi_profile *profile) {
+bool kanshi_switch(struct kanshi_state *state, struct kanshi_profile *profile,
+		kanshi_apply_done_func callback, void *data) {
 	struct kanshi_profile_output *matches[HEADS_MAX];
 	if (!match_profile(state, profile, matches)) {
 		return false;
 	}
 
-	apply_profile(state, profile, matches);
-	return true;
+	return apply_profile(state, profile, matches, callback, data);
 }
 
 static void output_manager_handle_done(void *data,
 		struct zwlr_output_manager_v1 *manager, uint32_t serial) {
 	struct kanshi_state *state = data;
 	state->serial = serial;
-	match_and_apply(state);
+	match_and_apply(state, NULL, NULL);
 }
 
 static void output_manager_handle_finished(void *data,
@@ -555,6 +592,8 @@ static void registry_handle_global(void *data, struct wl_registry *registry,
 		uint32_t bind_version = 2;
 		if (version >= 4) {
 			bind_version = 4;
+		} else if (version > bind_version) {
+			bind_version = version;
 		}
 		state->output_manager = wl_registry_bind(registry, name,
 			&zwlr_output_manager_v1_interface, bind_version);
@@ -617,17 +656,18 @@ static void destroy_config(struct kanshi_config *config) {
 	free(config);
 }
 
-bool kanshi_reload_config(struct kanshi_state *state) {
+bool kanshi_reload_config(struct kanshi_state *state,
+		kanshi_apply_done_func callback, void *data) {
 	fprintf(stderr, "reloading config\n");
 	struct kanshi_config *config = read_config(state->config_arg);
-	if (config != NULL) {
-		destroy_config(state->config);
-		state->config = config;
-		state->pending_profile = NULL;
-		state->current_profile = NULL;
-		return match_and_apply(state);
+	if (config == NULL) {
+		return false;
 	}
-	return false;
+	destroy_config(state->config);
+	state->config = config;
+	state->pending_profile = NULL;
+	state->current_profile = NULL;
+	return match_and_apply(state, callback, data);
 }
 
 static const char usage[] = "Usage: %s [options...]\n"
@@ -699,7 +739,10 @@ int main(int argc, char *argv[]) {
 
 	struct wl_registry *registry = wl_display_get_registry(display);
 	wl_registry_add_listener(registry, &registry_listener, &state);
-	wl_display_roundtrip(display);
+	if (wl_display_roundtrip(display) < 0) {
+		fprintf(stderr, "wl_display_roundtrip() failed\n");
+		return EXIT_FAILURE;
+	}
 
 	if (state.output_manager == NULL) {
 		fprintf(stderr, "compositor doesn't support "
